/*************************************************************************
 inner2.c 

 Modified version of routines in inner.c. These routines attempt to be more
 generic so as to be reusable in any context. In particular their
 application is in the controls for the outer controls. Eventually we
 really need to ensure that we use the same inner control for both the
 mesh estimates and the mesh estimates for the outer control control.
 *************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "blackScholes.h"
#include "control3.h"
#include "util.h"

/* weight macro now to replace teh weights array */
#define WEIGHT(time,prices_t,toPath) transProb(&(md->option.pData),prices_t,md->option.pData.price[time+1][toPath]) /md->distribution[(time)][(toPath)]


/*-----------------------------------------------------------------------
  The following set of routines are for the application of inner control 3.
  There are mainly four routines in this set:
  IControl3Exact       := computes exact value of control
  IControl3Estimate    := does a mesh estimate of control
  IControl3Variance    := computes a variance of the control mesh estimate
  IControl3Correlation := computes the correlation b/w the two

  Control3 is the max of two from a given set of five underliers. Hence 
  we use an approach from Stulz to evalute this. The actual formula is from
  Kolb. See inner.c.
*/

/*  IControl3Exact:
    Inputs: 
      time         := current time.
      prices_t     := prices at current time on the mesh.
      underliers   := number of underliers
      K            := strike price
      cov          := covariance matrix
      delta        := dividend rate
      r            := Risk-free interest rate
      tau          := time interval between sucessive time slices
      b            := number of mesh points
*/

double IControl3Exact(int time, double *prices_t, int underliers, double K,
 double **cov, double *delta, double r, double tau, int b){
  int max1, max2, i;
  double maxval1, maxval2;

  /* At the given time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(underliers, prices_t);
  maxval1 = prices_t[max1];

  for(i=0; i<underliers;i++){
    if(i != max1 && prices_t[i] > maxval2){
      max2 = i;
      maxval2 = prices_t[i];
    }
  }

/* Evaluate deflated value of option price as predicted by Stulz a la Kolb*/
  return(exp(-time*r)*
         maxOf2Risk(maxval1, maxval2, K, tau,
                    sqrt(cov[max1][max1]),
                    sqrt(cov[max2][max2]),
                    r, delta[max1], delta[max2],
                    cov[max1][max2]));


} /* end IControlExact */

   

/*  IControl3Exact:
    Inputs: 
      time         := current time.
      prices_t     := prices at current time on the mesh.
      prices_tp1   := prices at next time on each mesh point and underlier
      underliers   := number of underliers
      K            := strike price
      r            := Risk-free interest rate
      tau          := time interval between sucessive time slices
      b            := number of mesh points
*/

double IControl3Estimate(int time, double *prices_t, double **prices_tp1,
       double r, double K, double tau, int underliers, int b, 
       meshData *md){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* - - - - - At the current time find the max prices - - - - - */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(underliers, prices_t);
  maxval1 = prices_t[max1];

  for(i=0; i<underliers;i++){
    if(i != max1 && prices_t[i] > maxval2){
      max2 = i;
      maxval2 = prices_t[i];
    }
  }
  
  /* - - - - - Compute the mesh estimate now - - - - - */
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<b; toPath++){
    sumWeight += WEIGHT(time,prices_t,toPath);
    if( prices_tp1[toPath][max1] >=
        prices_tp1[toPath][max2]){
      maxPrice = prices_tp1[toPath][max1];
    }else{
      maxPrice = prices_tp1[toPath][max2];
    }
    if(maxPrice > K){
      deflatedPrice = exp(-r * tau)*
        (maxPrice - K);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices_t,toPath)*deflatedPrice;
  }
  estimate /= sumWeight;

  return(exp(-time*r)*estimate);
} /* End IControl3Estimate */


/*  IControl3Variance:
    Inputs: 
      time         := current time.
      prices_t     := prices at current time on the mesh.
      prices_tp1   := prices at next time on each mesh point and underlier
      underliers   := number of underliers
      K            := strike price
      r            := Risk-free interest rate
      tau          := time interval between sucessive time slices
      b            := number of mesh points
*/

double IControl3Variance(int time, double *prices_t, double **prices_tp1,
       double r, double K, double tau, int underliers, int b, meshData *md){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* - - - - - At the current time find the max prices - - - - - */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(underliers, prices_t);
  maxval1 = prices_t[max1];

  for(i=0; i<underliers;i++){
    if(i != max1 && prices_t[i] > maxval2){
      max2 = i;
      maxval2 = prices_t[i];
    }
  }
  
  /* - - - - - Compute the mesh estimate now - - - - - */
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<b; toPath++){
    if( prices_tp1[toPath][max1] >= prices_tp1[toPath][max2]){
      maxPrice = prices_tp1[toPath][max1];
    }else{
      maxPrice = prices_tp1[toPath][max2];
    }
    if(maxPrice > K){
      deflatedPrice = exp(-r * tau)* (maxPrice - K);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices_t,toPath)*deflatedPrice*deflatedPrice;
    sumWeight += WEIGHT(time,prices_t,toPath);
  }
  estimate /= sumWeight;

  return(exp(-2.0*time*r)*estimate);
} /* End IControl3Variance */

/*  IControl3Correlation:
    Inputs: 
      time         := current time.
      prices_t     := prices at current time on the mesh.
      prices_tp1   := prices at next time on each mesh point and underlier
      meshValue_tp1:= values at t+1 on the entire mesh.
      underliers   := number of underliers
      K            := strike price
      r            := Risk-free interest rate
      tau          := time interval between sucessive time slices
      b            := number of mesh points
*/

double IControl3Correlation(int time, double *prices_t, double **prices_tp1,
       double *meshValue_tp1, double r, double K, double tau, 
       int underliers, int b, meshData *md ){
  int  toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* - - - - - At the current time find the max prices - - - - - */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(underliers, prices_t);
  maxval1 = prices_t[max1];

  for(i=0; i<underliers;i++){
    if(i != max1 && prices_t[i] > maxval2){
      max2 = i;
      maxval2 = prices_t[i];
    }
  }
  
  /* - - - - - Compute the mesh estimate now - - - - - */
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<b; toPath++){
    if( prices_tp1[toPath][max1] >= prices_tp1[toPath][max2]){
      maxPrice = prices_tp1[toPath][max1];
    }else{
      maxPrice = prices_tp1[toPath][max2];
    }
    if(maxPrice > K){
      deflatedPrice = exp(-r * tau)* (maxPrice - K);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices_t,toPath)*exp(-time*r) *
                deflatedPrice*meshValue_tp1[toPath];
    sumWeight += WEIGHT(time,prices_t,toPath);
  }
  estimate /= sumWeight;

  return(estimate);
} /* End IControl3Correlation */

double IMeshEstimate(meshData *md, int time, double *prices_t, 
        double *meshValue_tp1){
  int max, toPath;
  double estimate;
  double w,v;

  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate +=  WEIGHT(time,prices_t,toPath)*meshValue_tp1[toPath];
  }
  return(estimate);
}

